/* System description file for Fujitsu UXP/V
   This file describes the parameters that system description files
   should define or not.  */

#include "usg5-4-2.h"

/* SYSTEM_TYPE should indicate the kind of system you are using.
 It sets the Lisp variable system-type.  */

#undef SYSTEM_TYPE
#define SYSTEM_TYPE "uxpv"
